import atexit
import datetime as dt
import json
import logging
import logging.config
import logging.handlers
import queue
import sys
from typing import override

LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


class JSONFormatter(logging.Formatter):
    def __init__(
            self,
            *,
            fmt_keys: dict[str, str] | None = None,
    ):
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    @override
    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord):
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.timezone.utc
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: msg_val
            if (msg_val := always_fields.pop(val, None)) is not None
            else getattr(record, val)
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        for key, val in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                if message.get("extra") is None:
                    message["extra"] = dict()
                message["extra"][key] = val

        return message


def get_stdout_logger(name: str) -> logging.Logger:
    json_formatter = JSONFormatter(fmt_keys={
        "level": "levelname",
        "message": "message",
        "timestamp": "timestamp",
        "logger": "name",
        "pathname": "pathname",
        "module": "module",
        "function": "funcName",
        "process_id": "process",
        "line": "lineno",
        "thread_name": "threadName"
    })
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(json_formatter)

    queue_handler_queue = queue.Queue()
    queue_handler = logging.handlers.QueueHandler(queue_handler_queue)
    queue_handler.listener = logging.handlers.QueueListener(queue_handler_queue, stdout_handler,
                                                            respect_handler_level=True)
    queue_handler.listener.start()
    atexit.register(queue_handler.listener.stop)

    logger = logging.getLogger(name)  # __name__ is a common choice
    logger.setLevel(logging.DEBUG)
    logger.handlers = [queue_handler]

    return logger
